# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

DS8ImgThread.py: Class and support functions for reading and processing the
                 images.

Last version: November 2021.
"""

from cv2 import (IMREAD_COLOR, resize, createMergeMertens, normalize,
                 NORM_MINMAX, createMergeDebevec, createCalibrateDebevec,
                 getRotationMatrix2D, warpAffine, imwrite, Laplacian,
                 IMWRITE_JPEG_QUALITY, IMWRITE_PNG_COMPRESSION, imdecode,
                 imencode, createTonemap, createTonemapReinhard, CV_64F,
                 createTonemapDrago, createTonemapMantiuk, cvtColor,
                 COLOR_BGR2GRAY, putText, FONT_HERSHEY_SIMPLEX)

from threading import Event

from struct import unpack, calcsize

from logging import info

from numpy import (uint8, shape, clip, ndarray, fromstring, bitwise_and,
                   percentile)

from io import BytesIO

from pathlib import Path

from PyQt5.QtCore import QThread, pyqtSignal

from exif import Image

from datetime import datetime

# Our own modules.

import config

from codes import NEW_IMAGE

# Class and support functions for the reading and treatment of images.
class imgThread(QThread):

    # Frame number indicator update signal.
    updateFrameNumSig = pyqtSignal(str)

    # Automatic exposure data update signal.
    updateAESig = pyqtSignal(int, float, float, float)

    # Exposure data update signal.
    updateSSSig = pyqtSignal(int, float, float, float)

    # A and D gains data update signal.
    updateGainsSig = pyqtSignal(int, int)

    # GUI status field refresh signal.
    updateStatusSig = pyqtSignal(str)

    # Image window refresh signal.
    displayImgSig = pyqtSignal(ndarray, str)

    # Histogram window refresh signal.
    plotHistogramSig = pyqtSignal(ndarray, str)

    # Engine stopped information signal.
    motorStoppedSig = pyqtSignal()

    # Application output signal reported by server.
    exitSig = pyqtSignal(bool)

    # Illumination status information signal.
    lightSig = pyqtSignal(str)

    # Last frame reached information signal.
    frameLimitReachedSig = pyqtSignal()

    # Exception on camera information signal.
    PiCameraExcpSig = pyqtSignal()

    # Displayed image information event.
    displayImgEvent = Event()
    displayImgEvent.set()

    # Displayed histogram information event.
    plotHistogramEvent = Event()
    plotHistogramEvent.set()

    def __init__(self, connection, app):
        QThread.__init__(self, parent=app)
        self.threadID = 1
        self.name = "ImgThread"
        self.conn = connection

        # Temporary storage of the image received from the server.
        self.image_stream = BytesIO()

        # Image size received from server.
        self.image_len = None

        # Opencv image.
        self.cvimg = ndarray

        # Group of images from multiple exposures to obtain an HDR image.
        self.imglist = []

        # Index used for the exposure time matrix.
        self.indexETM = 0

        # Indicator of the type of information received from the server.
        self.imgflag = ""

        # Exposure time reported by camera.
        self.exposureTime = 0

        # Generate angle round masks.
        self.roundcornTLImg = config.readImgFromQrc(":/roundcornTL.png")
        self.roundcornTRImg = config.readImgFromQrc(":/roundcornTR.png")
        self.roundcornBRImg = config.readImgFromQrc(":/roundcornBR.png")
        self.roundcornBLImg = config.readImgFromQrc(":/roundcornBL.png")

        # Image name.
        self.imageName = ""

        # Image file name.
        self.filename = ""

        # Permission to save bracketing images.
        self.saveBracketPerm = False

        # Focus index coordinates.
        self.x1 = 0
        self.y1 = self.x1
        self.x2 = 0
        self.y2 = 0

        # Final image resolution.
        self.finalh = config.imgCapFinalH
        self.finalw = config.imgCapFinalW

    # Sending information to the status field of the GUI.
    def updateStatus(self, status):
        self.updateStatusSig.emit(status)

    # Rotating and cropping the image.
    def postProcess(self, img):

        h, w = img.shape[:2]

        # Image rotation if selected.
        if config.rotation:
            #info("Rotating image")
            img = self.imageRotation(img, w, h)

        # Cropping the image if selected.
        if config.cropping:
            #info("Cropping image")
            img = img[config.cropT:h - config.cropB,
                      config.cropL:w - config.cropR]

        return img

    # Image rotation.
    def imageRotation(self, img, w, h):
        rotMtx = getRotationMatrix2D((w / 2, h / 2), config.rotationValue, 1)
        img = warpAffine(img, rotMtx, (w, h))

        #info("Rotated image")

        return img

    # Scaling the image to the maximum dimensions specified in the config.py
    # file.
    def imageResize(self, img):
        h, w = img.shape[:2]

        config.imgCapIniH = h
        config.imgCapIniW = w

        self.finalh = config.imgCapFinalH
        self.finalw = int((self.finalh / h) * w)
        if self.finalw > config.imgCapFinalW:
            self.finalw = config.imgCapFinalW
            self.finalh = int((self.finalw / w) * h)
        img = resize(img, (self.finalw, self.finalh))

        #info("Resized image")

        return img

    # Draw the sharpness index on the image.
    def drawImageSharpness(self, img):

        self.x1 = int(config.imgCapIniH / 10.8)
        self.y1 = self.x1
        self.x2 = int(config.imgCapIniH / 3.4)
        self.y2 = int(config.imgCapIniH / 7)
        self.fontScale = round(config.imgCapIniH / 540, 1)
        self.tickness = int(self.fontScale * 2)

        imgSharp = self.imageSharpness(img)

        config.numMeasSharp += 1

        putText(img, "Focus: ", (self.x1, self.y1),
                FONT_HERSHEY_SIMPLEX, self.fontScale, (0, 0, 255),
                self.tickness)

        putText(img, str(imgSharp), (self.x2, self.y1),
                FONT_HERSHEY_SIMPLEX, self.fontScale, (0, 0, 255),
                self.tickness)

        if config.numMeasSharp > config.valSharp:

            if (imgSharp > config.maxSharpness):
                config.maxSharpness = imgSharp

            putText(img, "Max: ", (self.x1, self.y2),
                    FONT_HERSHEY_SIMPLEX, self.fontScale, (0, 0, 255),
                    self.tickness)

            putText(img, str(config.maxSharpness), (self.x2, self.y2),
                    FONT_HERSHEY_SIMPLEX, self.fontScale, (0, 0, 255),
                    self.tickness)

        return img

    # Calculation of the image sharpness index.
    def imageSharpness(self, img):
        imgSharp = 0
        imgSharp = Laplacian(cvtColor(img, COLOR_BGR2GRAY), CV_64F).var()
        imgSharp = round(imgSharp, 2)

        return imgSharp

    # Rounding the angles of the image.
    def roundCorners(self, img):
        h, w = img.shape[:2]

        img[0:50, 0:50] = bitwise_and(img[0:50, 0:50], self.roundcornTLImg)
        img[0:50, w - 50:w] = bitwise_and(img[0:50, w - 50:w],
                                          self.roundcornTRImg)
        img[h - 50:h, w - 50:w] = bitwise_and(img[h - 50:h, w - 50:w],
                                              self.roundcornBRImg)
        img[h - 50:h, 0:50] = bitwise_and(img[h - 50:h, 0:50],
                                          self.roundcornBLImg)
        return img

    # Merging bracketing images to obtain an HDR image.
    def blendImgList(self):

        if config.blender == "Mertens":
            blender = createMergeMertens()
            img = blender.process(self.imglist)

            # The normalize function adjusts the values from the above process
            # between 0 and 1. As a result, the contrast is significantly
            # reduced.
            # img = normalize(img, None, 0., 1., NORM_MINMAX)

            # Function proposed by Rolf Henkel (cpixip) to carry out the
            # normalization.
            # Percentiles are applied to discard the brightest and darkest
            # pixels in the image.
            minimum = percentile(img, config.MertPercLow)
            maximum = percentile(img, config.MertPercHigh)
            scaler = 1.0 / (maximum - minimum + 1e-6)

            img = scaler * (img - minimum)

        else:
            # In tests carried out, it has been found that, to obtain good
            # results with the Debevec algorithm, it is required to take enough
            # images. Minimum 6 images

            # Get the response function of the camera (CRF).
            calibrateDebevec = createCalibrateDebevec()
            responseDebevec = calibrateDebevec.process(self.imglist,
                                                       config.exposureTimes)

            # Merge the images into a linear HDR image.
            blender = createMergeDebevec()
            hdrDebevec = blender.process(self.imglist, config.exposureTimes,
                                         responseDebevec)

            # Apply tone mapping.
            if config.toneMap == "Simple":
                img = self.toneMapSimple(hdrDebevec)
            elif config.toneMap == "Reinhard":
                img = self.toneMapReinhard(hdrDebevec)
            elif config.toneMap == "Drago":
                img = self.toneMapDrago(hdrDebevec)
            elif config.toneMap == "Mantiuk":
                img = self.toneMapMantiuk(hdrDebevec)

        # We convert to BGR matrix.
        img = clip(img * 255, 0, 255).astype('uint8')

        info("Images fusion done")

        return img

    # Apply simple tone mapping.
    def toneMapSimple(self, hdrDebevec):

        toneMap = createTonemap(gamma=config.SimpleGamma)
        img = toneMap.process(hdrDebevec)
        return img

    # Apply Reinhard method tone mapping.
    def toneMapReinhard(self, hdrDebevec):

        toneMap = createTonemapReinhard(config.ReinhardGamma,
                                        config.ReinhardIntensity,
                                        config.ReinhardLight,
                                        config.ReinhardColor)

        img = toneMap.process(hdrDebevec)
        return img

    # Apply Drago method tone mapping.
    def toneMapDrago(self, hdrDebevec):

        toneMap = createTonemapDrago(config.DragoGamma,
                                     config.DragoSaturation,
                                     config.DragoBias)
        img = toneMap.process(hdrDebevec)
        img *= 3
        return img

    # Apply Mantiuk method tone mapping.
    def toneMapMantiuk(self, hdrDebevec):

        toneMap = createTonemapMantiuk(config.MantiukGamma,
                                       config.MantiukScale,
                                       config.MantiukSaturation)
        img = toneMap.process(hdrDebevec)
        img *= 3
        return img

    # Show captured image.
    def showImage(self, img, title=""):
        self.displayImgEvent.wait(10)
        self.displayImgSig.emit(img, title)
        self.displayImgEvent.clear()

    # Show histogram of captured image.
    def showHist(self, img, title=""):
        self.plotHistogramEvent.wait(10)
        self.plotHistogramSig.emit(img, title)
        self.plotHistogramEvent.clear()

    # This function is used to name the test files.
    def testFileName(self):

        i = 1
        while True:
            # self.imageName = "Test{:05d}.png".format(i)
            # self.imageName = "Test{:05d}.tif".format(i)
            self.imageName = "Test{:05d}.jpg".format(i)

            self.filename = config.capFolder.strip() + "/" + self.imageName
            if Path(self.filename).exists():
                i += 1
                continue
            else:
                break

    def finalizeImage(self, img):

        imgSharp = 0

        if config.lastMode == "P":
            self.imageName = "Previous " + str(config.numImgRec)

        elif config.lastMode == "T":
            self.testFileName()

        elif config.lastMode == "C":
            # PNG image format has been tested. It is very slow.
            # self.imageName = "img{:05d}.png".format(config.file_number)
            # self.imageName = "img{:05d}.tif".format(config.file_number)
            self.imageName = "img{:05d}.jpg".format(config.file_number)

        # Sharpness index calculation and display.
        # Only for preview images.
        if (config.lastMode == "P" and config.showSharp
                and config.motorNotMoving):

            img = self.drawImageSharpness(img)

        # Image resize.
        img = self.imageResize(img)

        # The image histogram is displayed.
        if config.showHist:
            self.showHist(img, self.imageName)

        # Rounding the image angles.
        if config.roundcorns:
            img = self.roundCorners(img)

        # The image is shown.
        self.showImage(img, self.imageName)

        return img

    # Function to save bracketing image files.
    def writeBracketImgFile(self, img):

        if config.lastMode == "T":
            self.testFileName()

            filename = (self.filename[:-4]
                        + "-{:02d}.jpg".format(self.indexETM + 1))

            filename = Path(filename)

        else:
            filename = (Path(config.capFolder.strip() +
                             "/img{:05d}-{:02d}.jpg".format(config.file_number,
                                                              self.indexETM + 1)))

        # We encode img in jpg.
        (ret, img_jpg) = imencode(".jpg", img, (int(IMWRITE_JPEG_QUALITY), 97))

        img_jpg = img_jpg.tobytes()

        # We add exif information.
        img_exif = Image(img_jpg)

        img_exif.datetime = datetime.today().strftime("%d-%m-%Y %H:%M:%S")
        img_exif.make = "Raspberry Pi"
        img_exif.model = config.cameraType + " Camera"
        img_exif.exposure_time = self.exposureTime

        # The image is saved.
        with (open(filename, "wb")) as imfile:
            imfile.write(img_exif.get_file())

    def writeImgFile(self, img):

        self.filename = config.capFolder.strip() + "/" + self.imageName

        with open(Path(self.filename), "w"):
            # Write PNG file.
            # imwrite(self.filename, img,
            #        [int(IMWRITE_PNG_COMPRESSION), 5])

            # Write TIFF file.
            # imwrite(self.filename, img)

            # Write JPG file.
            imwrite(self.filename, img,
                    [int(IMWRITE_JPEG_QUALITY), 97])

        if config.testImg:
            info("Test image saved in: " + str(self.filename))

        else:
            info("Captured image saved in: " + str(self.filename))

    # With this function the server is requested to capture and send a new
    # image.
    def newImage(self):
        config.ctrl_conn.write(NEW_IMAGE + "\n")
        config.ctrl_conn.flush()
        info("Image " + str(config.numImgRec) + " requested")

    # Processing functions of the information received from the server.

    # Server reports exception on camera.
    # Flag d -> discard multiple images previously received while waiting for
    # new images.
    def imgflag_d(self):
        self.imglist = []
        self.indexETM = 0
        self.PiCameraExcpSig.emit()
        info("Server reports exception in camera. Discarded images")

    # Flag e -> automatic exposure time.
    def imgflag_e(self):
        ssAE = unpack("<l", self.conn.read(calcsize("<l")))[0]
        again = unpack("<f", self.conn.read(calcsize("<f")))[0]
        dgain = unpack("<f", self.conn.read(calcsize("<f")))[0]
        framerate = unpack("<f", self.conn.read(calcsize("<f")))[0]
        self.updateAESig.emit(ssAE, again, dgain, framerate)

    # Flag f -> exposure time and analog and digital gains.
    def imgFlag_f(self):
        ss = unpack("<l", self.conn.read(calcsize("<l")))[0]
        again = unpack("<f", self.conn.read(calcsize("<f")))[0]
        dgain = unpack("<f", self.conn.read(calcsize("<f")))[0]
        framerate = unpack("<f", self.conn.read(calcsize("<f")))[0]
        self.updateSSSig.emit(ss, again, dgain, framerate)

    # Flag g -> blue and red gains.
    def imgFlag_g(self):
        gblue = unpack("<i", self.conn.read(calcsize("<i")))[0]
        gred = unpack("<i", self.conn.read(calcsize("<i")))[0]
        self.updateGainsSig.emit(gblue, gred)

    # Flags l -> light on and L -> light off.
    def imgFlag_lL(self):
        self.lightSig.emit(self.imgflag)

    def imgFlag_spab(self):
        self.exposureTime = unpack("<i", self.conn.read(calcsize("<i")))[0]
        self.image_len = unpack("<L", self.conn.read(calcsize("<L")))[0]

        # Save image data to temporary storage.
        self.image_stream.write(self.conn.read(self.image_len))
        self.image_stream.seek(0)

        # File to opencv image.
        self.cvimg = imdecode(fromstring(self.image_stream.read(self.image_len),
                                         dtype=uint8), IMREAD_COLOR)

    # Preview Image.
    def imgFlag_p(self):

        if type(self.cvimg) != ndarray:
            return

        info("Preview image " + str(config.numImgRec) + " received"
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        config.numImgRec += 1

        # New image is requested from the server.
        if config.prevOn:
            self.newImage()

        self.cvimg = self.postProcess(self.cvimg)
        self.cvimg = self.finalizeImage(self.cvimg)

    # Digitized frame with a single image.
    def imgFlag_s(self):

        info("Single image " + str(config.numImgRec) + " received"
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        config.numImgRec += 1

        # New image is requested from the server.
        if (config.captureOn and config.file_number < config.frame_limit):
            self.newImage()

        self.cvimg = self.postProcess(self.cvimg)
        self.cvimg = self.finalizeImage(self.cvimg)
        self.writeImgFile(self.cvimg)

        if config.lastMode == "C":

            if config.file_number >= config.frame_limit:
                self.frameLimitReachedSig.emit()
            config.file_number += 1

    # Digitized frame with multiple images merged.
    # One of several bracketing images.
    def imgFlag_a(self):

        # Permission to save bracketing images.
        if config.saveBracketImg and not self.indexETM:
            self.saveBracketPerm = True

        if not config.saveBracketImg and not self.indexETM:
            self.saveBracketPerm = False

        info("Bracketing image " + str(config.numImgRec) + " received"
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        self.cvimg = self.postProcess(self.cvimg)

        # The image is saved in a list to later perform the fusion of the
        # images.

        self.imglist.append(self.cvimg)

        # The HDR Debevec algorithm uses time in s.
        self.exposureTime = float(self.exposureTime * 1e-6)
        config.exposureTimes[self.indexETM] = self.exposureTime

        # The bracketing image is saved if this option is selected.

        if self.saveBracketPerm:
            self.writeBracketImgFile(self.cvimg)

        # The index is increased.

        self.indexETM += 1

    # The last of several bracketing images.
    def imgFlag_b(self):

        info("Last bracketing image " + str(config.numImgRec) + " received"
             + " - " + str(self.image_len) + " bytes - Exp. time "
             + str(self.exposureTime) + " us")

        config.numImgRec += 1

        # New image is requested from the server.
        if config.captureOn and (config.file_number < config.frame_limit):
            self.newImage()

        self.cvimg = self.postProcess(self.cvimg)

        # The image is saved in a list to later perform the fusion of the
        # images.
        self.imglist.append(self.cvimg)

        # The HDR Debevec algorithm uses time in s.
        self.exposureTime = float(self.exposureTime * 1e-6)
        config.exposureTimes[self.indexETM] = self.exposureTime

        # The bracketing image is saved if this option is selected.

        if self.saveBracketPerm:
            self.writeBracketImgFile(self.cvimg)

        self.cvimg = self.blendImgList()
        self.cvimg = self.finalizeImage(self.cvimg)
        self.writeImgFile(self.cvimg)

        # Cleaning the list of images and exposure times.
        self.imglist = []
        self.indexETM = 0

        if config.lastMode == "C":

            if config.file_number >= config.frame_limit:
                self.frameLimitReachedSig.emit()
            config.file_number += 1

    # Imaging thread main loop.

    def run(self):
        info("Executing main function of image thread")

        while not config.exitFlag:

            # The flag of the information sent by the server is obtained.
            self.imgflag = self.conn.read(1)
            try:
                self.imgflag = self.imgflag.decode()
            except:
                continue

            # Treatment of data sent by the server.

            # Server reports exception on camera.
            # Flag d -> discard bracketing images previously received while
            # waiting for new images.
            if self.imgflag == "d":
                self.imgflag_d()

            # Flag e -> automatic exposure time.
            elif self.imgflag == "e":
                self.imgflag_e()

            # Flag f -> exposure time and analog and digital gains.
            elif self.imgflag == "f":
                self.imgFlag_f()

            # Flag g -> red and blue gains.
            elif self.imgflag == "g":
                self.imgFlag_g()

            # Flags l -> light on and L -> light off
            elif self.imgflag == "l" or self.imgflag == "L":
                self.imgFlag_lL()

            # Flag m -> motor stopped.
            elif self.imgflag == "m":
                self.motorStoppedSig.emit()

            # Flag c -> frame advance
            # Flag C -> reverse of a frame
            elif self.imgflag == "c":
                self.updateFrameNumSig.emit("c")

            elif self.imgflag == "C":
                self.updateFrameNumSig.emit("C")

            # Flag X -> application output reported by server.
            elif self.imgflag == "X":
                self.exitSig.emit(True)

            # Flag p -> preview image.
            # Flag s -> single capture image.
            # Flag a -> image from a series of images to merge.
            # Flag b -> last image of the serie.

            else:

                if (self.imgflag == "p" or self.imgflag == "s" or
                        self.imgflag == "a" or self.imgflag == "b"):
                    self.imgFlag_spab()

                # Preview image.
                if self.imgflag == "p":
                    self.imgFlag_p()

                # Digitized frame with a single image.
                if self.imgflag == "s":
                    self.imgFlag_s()

                # Digitized frame with multiple images merged.
                # One of several bracketing images.
                if self.imgflag == "a":
                    self.imgFlag_a()

                # The last of several bracketing images.
                if self.imgflag == "b":
                    self.imgFlag_b()

                # Cleaning the temporary storage of the image received from the
                # server.
                self.image_stream.seek(0)
                self.image_stream.truncate()

        self.conn.close()
        info("Imaging thread " + self.name + " finalized")


