#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
DSuper8 project based on Joe Herman's rpi-film-capture.

Software modified by Manuel Ángel.

User interface redesigned by Manuel Ángel.

DSuper8.py: Main client software program.

Last version: November 2021.
"""

from socket import socket, AF_INET, SOCK_STREAM

from pathlib import Path

from sys import argv, stdout, exit

from logging import INFO, basicConfig, info

from configparser import Error

from PyQt5.QtWidgets import QApplication, QStyleFactory

from PyQt5.QtCore import QFile, QIODevice


# Our own modules.

# Resource file: default configuration file, splash and rounding masks of image
# angles.
import qrc_DSuper8

# The global variables of the modules are defined here.
import config

# Classes for the definition of our UI and for the image window.
from DS8Dialogs import DS8Dialog, DS8ImgDialog

# Class and support functions for the reading and treatment of images.
from DS8ImgThread import imgThread

# Creation of 2 connections: one for sending control sequences and another for
# receiving image data.


def setup_conns(image_socket, control_socket):
    try:
        image_socket.connect((config.server_ip, 8000))
        control_socket.connect((config.server_ip, 8001))
        img_conn = image_socket.makefile("rb")
        ctrl_conn = control_socket.makefile("w")
        info("Connections with the server established")
        return (img_conn, ctrl_conn)

    except:
        info("It seems that the server is not working")
        info("The server must be running before the client program starts")
        image_socket.close()
        control_socket.close()
        exit()


if __name__ == "__main__":
    # Severity level of the log set to INFO.
    basicConfig(stream=stdout, level=INFO, format="%(asctime)s - %(levelname)s "
                + "- %(message)s", datefmt="%Y-%m-%d %H:%M:%S")

    app = QApplication(argv)
    app.setStyle(QStyleFactory.create("Fusion"))

    configError = False
    config.testImg = True

    # Establishing connections with the server.
    image_socket = socket(AF_INET, SOCK_STREAM)
    control_socket = socket(AF_INET, SOCK_STREAM)
    (img_conn, ctrl_conn) = setup_conns(image_socket, control_socket)
    config.ctrl_conn = ctrl_conn

    # Thread for receiving and processing the images sent by the server.
    imgthread = imgThread(img_conn, app)

    # Graphical user interface.
    winUI = DS8Dialog()

    # By default, the user interface is designed to use the HQ camera.
    # If the Raspberry Pi V1 camera is used, the UI must be modified.
    if config.cameraType == "V1":
        winUI.V1CamGUI()

    # Loading the initial configuration.

    if Path(config.configfile.strip()).is_file():
        try:
            winUI.config.read([Path(config.configfile.strip())])
            if not winUI.config.validateConfigFile():
                raise Error
            winUI.config.updateUIfromConfig(winUI)
            winUI.configFileBox.setText(config.configfile)
            winUI.saveConfigButton.setEnabled(True)
            info(config.configfile + " configuration file loaded")

        except Error:
            configError = True
            info(config.configfile + " not a valid configuration file")

    else:
        configError = True
        info("Configuration file " + config.configfile + " does not exist")
        info("Modify path to the configuration file in config.py of the client"
             " program")

    if configError:
        winUI.configFileBox.setText("")
        winUI.saveConfigButton.setEnabled(False)
        info("Default configuration is loaded")
        confFile = QFile(":/DSuper8Default.conf")
        confFile.open(QIODevice.ReadOnly | QIODevice.Text)
        confText = confFile.readAll()
        confText = str(confText, encoding="utf-8")
        try:
            winUI.config.read_string(confText)
            winUI.config.updateUIfromConfig(winUI)
        except Error:
            info("Could not load default configuration")
            info("Manually configure UI options")
        confFile.close()

    # Image window.
    winImg = DS8ImgDialog(winUI)
    splash = winUI.lastShowImg
    winImg.displayImg(splash, "Splash")
    winImg.resize(config.imgWinWidth, config.imgWinHeight)

    # Slot configuration to allow the image sequence to update the image window
    # and user interface.
    winImg.setupThreadingUpdates(imgthread)
    winUI.setupThreadingUpdates(imgthread)

    # Image thread start.
    imgthread.start()
    info("Images thread started")

    # Initial server configuration to match the settings defined in the user
    # interface.
    winUI.sendInitConfig()

    # Enable/Disable controls to match settings loaded in initial setup.
    winUI.initPostCapture()

    # Show windows.
    winUI.show()
    winImg.show()
    info("Windows displayed")

    # Open histogram window if selected.
    if winUI.showHist.isChecked():
        winUI.initHistogram(True)

    winUI.updateStatus("Program started")

    rtnVal = app.exec_()
    info("Windows closed")
    image_socket.close()
    control_socket.close()
    info("Server connections closed")
    info("Finalized")
    exit(rtnVal)
